'''title'''
__title__ = 'musicdl'
'''description'''
__description__ = 'Musicdl: A lightweight music downloader written in pure python'
'''url'''
__url__ = 'https://github.com/CharlesPikachu/musicdl'
'''version'''
__version__ = '2.8.4'
'''author'''
__author__ = 'Zhenchao Jin'
'''email'''
__email__ = 'charlesblwx@gmail.com'
'''license'''
__license__ = 'Apache License 2.0'
'''copyright'''
__copyright__ = 'Copyright 2020-2030 Zhenchao Jin'